/*
 * topological-hex, a program to compute combinatorial hexahedral meshes.
 *
 * Copyright (C) <2018> <Université catholique de Louvain (UCL), Belgique>
 *
 * List of the contributors to the development, description and complete
 * License: see LICENSE file.
 *
 * This program (topological-hex) is free software:
 * you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see COPYING file).  If not,
 * see <http://www.gnu.org/licenses/>.
 */


#include <string.h>
#include <stdio.h>

#include "dfs2d.h"

typedef struct fixed_point_delta2d {
  uint32_t new_bound_mask, diagonal_mask;
  ALIGNED vertex_set domain_changes[2];

  uint32_t num_occurrences_changes, max_edge_change;
  edge_index occurred[3];

  ALIGNED vertex_set removed_from[4];
  ALIGNED vertex_set allowed_set_changes[4];
  ALIGNED vertex_set new_neighbors[4];
} fixed_point_delta2d;

static
void start_quad(solver2d *solver, quad_index quad_id);

static
void choose_quad_vertex(solver2d *solver, quad_index quad_id, uint8_t i);

static
edge_index pop_edge(solver2d *solver, uint32_t *delta_size);

static
void undo_pop_edge(solver2d *solver, edge_index edge, uint32_t delta_size);

static
bool fixed_point2d(solver2d *solver, quad_index quad_id,
                   fixed_point_delta2d *delta);

static
void undo_fixed_point2d(solver2d *solver, quad_index quad_id,
                        const fixed_point_delta2d *delta);

void solver2d_run(solver2d *solver, solver2d_callback cb, void *user_data) {
  solver->cb        = cb;
  solver->user_data = user_data;

  solver->num_branches = 0;

  start_quad(solver, 0);
}

static
void start_quad(solver2d *solver, quad_index quad_id) {
  uint32_t delta_queue;
  edge_index edge = pop_edge(solver, &delta_queue);

  if (edge == NO_EDGE) {
    vertex_index v;
    for (v = 0; v < solver->num_vertices; v++) {
      if (solver->valences[v] != 0)
        break;
    }

    for (; v < solver->num_vertices; v++) {
      if (solver->valences[v] != 4)
        break;
    }

    if (v == solver->num_vertices) {
      solver->cb(solver, quad_id, solver->user_data);
    }
  }
  else if (quad_id != solver->num_quads) {
    memcpy(solver->quads + 4 * quad_id, solver->edges + 2*edge,
           2*sizeof(*solver->quads));

    vertex_index u = solver->quads[4*quad_id + 0];
    vertex_index v = solver->quads[4*quad_id + 1];

    if (solver->valences[u] > 0 && solver->valences[v] > 0) {
      solver->valences[u]--;
      solver->valences[v]--;

      for (uint32_t i = 0; i < 2; i++) {
        solver->domains[i] =
          solver->allowed_neighbors[solver->quads[4*quad_id + i]];
        vertex_set_subtract(
          solver->domains + i,
          solver->known_diagonals + solver->quads[4*quad_id + (i^1)]);
        for (uint32_t j = 0; j < 2; j++)
          vertex_set_remove(solver->domains + i, solver->quads[4*quad_id + j]);
      }

      solver->bound_mask = 0;

      fixed_point_delta2d trail;
      if (fixed_point2d(solver, quad_id, &trail))
        choose_quad_vertex(solver, quad_id, 0);
      undo_fixed_point2d(solver, quad_id, &trail);

      solver->valences[u]++;
      solver->valences[v]++;
    }
  }

  undo_pop_edge(solver, edge, delta_queue);
}

static
void choose_quad_vertex(solver2d *solver, quad_index quad_id, uint8_t i) {
  if (i == 2)  {
    uint32_t bound_mask = solver->bound_mask;

    ALIGNED vertex_set domains[2];
    memcpy(domains, solver->domains, sizeof(domains));
    start_quad(solver, quad_id + 1);
    memcpy(solver->domains, domains, sizeof(domains));

    solver->bound_mask = bound_mask;
  }
  else {
    if (solver->bound_mask & (1 << i))
      choose_quad_vertex(solver, quad_id, i + 1);
    else {
      vertex_set old_domain = solver->domains[i];

      vertex_set restricted_domain = old_domain;

      for (uint32_t i = 0;
           i < sizeof(solver->for_valence)/sizeof(*solver->for_valence); i++) {
        if (!solver->for_valence[i].used)
          continue;

        vertex_set to_remove;
        vertex_set_init_up_to(&to_remove, solver->for_valence[i].limit);

        vertex_set allowed;
        vertex_set_init_up_to(
          &allowed,
          solver->for_valence[i].max_vertex_id ==
          solver->for_valence[i].limit ?
          solver->for_valence[i].max_vertex_id :
          solver->for_valence[i].max_vertex_id + 1);

        vertex_set_subtract(&to_remove, &allowed);
        vertex_set_subtract(&restricted_domain, &to_remove);
      }

      uint32_t other_i = i ^ 1;
      if (solver->bound_mask & (1 << other_i))
        vertex_set_intersect(&restricted_domain,
                             solver->allowed_neighbors +
                             solver->quads[4*quad_id + 2+other_i]);

      solver->num_branches++;

      VERTEX_SET_FOREACH(restricted_domain, v) {
        vertex_set_init_empty(solver->domains + i);
        vertex_set_insert(solver->domains + i, v);

        uint32_t valence = solver->original_valences[v];
        bool increased = false;

        if (v >= solver->for_valence[valence].max_vertex_id) {
          increased = true;
          solver->for_valence[valence].max_vertex_id++;
        }

        fixed_point_delta2d trail;
        if (fixed_point2d(solver, quad_id, &trail))
          choose_quad_vertex(solver, quad_id, i + 1);
        undo_fixed_point2d(solver, quad_id, &trail);

        if (increased)
          solver->for_valence[valence].max_vertex_id--;
      }

      solver->domains[i] = old_domain;
    }
  }
}

static
edge_index pop_edge(solver2d *solver, uint32_t *delta_size) {
  uint32_t old_offset = solver->edge_queue_offset;
  while (solver->edge_queue_offset < solver->edge_queue_size &&
         solver->num_occurrences[solver->edge_queue_offset] == 0) {
    solver->edge_queue_offset++;
  }

  *delta_size = solver->edge_queue_offset - old_offset;
  if (solver->edge_queue_offset == solver->edge_queue_size)
    return NO_EDGE;

  (*delta_size)++;
  edge_index edge = solver->edge_queue[solver->edge_queue_offset++];
  solver->num_occurrences[edge]--;
  return edge;
}

static
void undo_pop_edge(solver2d *solver, edge_index edge, uint32_t delta_size) {
  solver->edge_queue_offset -= delta_size;
  if (edge != NO_EDGE)
    solver->num_occurrences[edge]++;
}

static bool finish_quad_diagonal(solver2d *solver, fixed_point_delta2d *delta,
                                 quad_index quad_id,
                                 vertex_index u, vertex_index v,
                                 quad_vertex_index i);

static bool finish_edge2d(solver2d *solver, fixed_point_delta2d *delta,
                          quad_vertex_index i, quad_vertex_index j,
                          vertex_index u, vertex_index v);

static void forbid_edge2d(solver2d *solver, fixed_point_delta2d *delta,
                          quad_vertex_index i, quad_vertex_index j,
                          vertex_index a, vertex_index b);

static
bool fixed_point2d(solver2d *solver, quad_index quad_id,
                   fixed_point_delta2d *delta) {
  bool failed = false;

  uint32_t bound_mask = solver->bound_mask;

  for (uint32_t i = 0; i < 2; i++) {
    if ((solver->bound_mask & (1 << i)) == 0 &&
        vertex_set_unit(&solver->domains[i])) {
      bound_mask |= (1 << i);
      solver->quads[4*quad_id + 2+i] = vertex_set_min(solver->domains + i);
    }
    else if (vertex_set_empty(solver->domains + i))
      failed = true;
  }

  delta->diagonal_mask = 0;

  for (uint32_t i = 0; i < 2; i++)
    vertex_set_init_empty(delta->domain_changes + i);

  delta->max_edge_change = 0;
  delta->num_occurrences_changes = 0;

  for (quad_vertex_index i = 0; i < 4; i++)
    vertex_set_init_empty(delta->removed_from + i);
  for (quad_vertex_index i = 0; i < 4; i++)
    vertex_set_init_empty(delta->allowed_set_changes + i);
  for (quad_vertex_index i = 0; i < 4; i++)
    vertex_set_init_empty(delta->new_neighbors + i);

  uint32_t to_visit = bound_mask & ~solver->bound_mask;

  while (!failed && to_visit != 0) {
    uint32_t i = find_lsb32(to_visit);
    to_visit &= ~(1 << i);

    uint32_t other_i = i ^ 1;

    vertex_index v = solver->quads[4*quad_id + 2+i];
    vertex_index below_v = solver->quads[4*quad_id + i];
    vertex_index other_v = solver->quads[4*quad_id + 2+other_i];
    vertex_index below_other_v = solver->quads[4*quad_id + other_i];

    if (solver->valences[v]-- == 0) {
      failed = true;
      break;
    }

    if (!finish_quad_diagonal(solver, delta, quad_id, v, below_other_v, i)) {
      failed = true;
      break;
    }

    forbid_edge2d(solver, delta, 2+i, other_i, v, below_other_v);

    if (!finish_edge2d(solver, delta, 2+i, i, v, below_v)) {
      failed = true;
      break;
    }

    if (bound_mask & (1 << other_i)) {
      if (!finish_edge2d(solver, delta, 2+i, 2+other_i, v, other_v)) {
        failed = true;
        break;
      }
    }

    if (vertex_set_remove(solver->domains + other_i, v))
      vertex_set_insert(delta->domain_changes + other_i, v);

    for (uint32_t j = 0; j < 2; j++) {
      uint32_t other_j = j ^ 1;

      vertex_set to_keep = solver->domains[j];

      if ((bound_mask & (1 << other_j)) != 0) {
        vertex_set_intersect(&to_keep,
                             solver->allowed_neighbors +
                             solver->quads[4*quad_id + 2+other_j]);
      }

      vertex_set_intersect(&to_keep,
                           solver->allowed_neighbors +
                           solver->quads[4*quad_id + j]);

      vertex_set removed = solver->domains[j];
      vertex_set_subtract(&removed, &to_keep);
      vertex_set_union(delta->domain_changes + j, &removed);
      solver->domains[j] = to_keep;

      if ((bound_mask & (1 << j)) == 0 &&
          vertex_set_unit(solver->domains + j)) {
        bound_mask |= (1 << j);
        to_visit |= (1 << j);
        solver->quads[4*quad_id + 2+j] = vertex_set_min(solver->domains + j);
      }
      else if (vertex_set_empty(solver->domains + j)) {
        failed = true;
        break;
      }
    }
  }

  delta->new_bound_mask = bound_mask & ~solver->bound_mask;
  solver->bound_mask = bound_mask;

  if ((solver->bound_mask & ((1 << 2)-1)) == (1 << 2)-1) {
    for (uint32_t i = solver->edge_queue_offset;
         i < solver->edge_queue_size; i++) {
      edge_index e = solver->edge_queue[i];
      if (solver->num_occurrences[e] == 0) continue;

      if (solver->valences[solver->edges[2*e + 0]] == 0 ||
          solver->valences[solver->edges[2*e + 1]] == 0) {
        failed = true;
        break;
      }
    }
  }

  return !failed;
}

static
void undo_fixed_point2d(solver2d *solver, quad_index quad_id,
                        const fixed_point_delta2d *delta) {
  solver->bound_mask &= ~delta->new_bound_mask;

  for (uint32_t i = 0; i < delta->num_occurrences_changes; i++) {
    edge_index e = delta->occurred[i];
    solver->num_occurrences[e]++;
    if (solver->num_occurrences[e] == 2)
      solver->edge_queue_size--;
  }

  for (edge_index i = solver->max_edge_id - delta->max_edge_change;
       i < solver->max_edge_id; i++) {
    vertex_index u = solver->edges[2*i + 0];
    vertex_index v = solver->edges[2*i + 1];
    solver->edge_map[diagonal_index(solver->num_vertices, u, v)] = NO_EDGE;
  }

  solver->max_edge_id -= delta->max_edge_change;

  for (quad_vertex_index i = 0; i < 4; i++) {
    if (i >= 2) {
      vertex_set_union(solver->domains + (i - 2),
                       delta->domain_changes + (i - 2));

      if (delta->new_bound_mask & (1 << (i-2)))
        solver->valences[solver->quads[4*quad_id + i]]++;
    }

    VERTEX_SET_FOREACH(delta->removed_from[i], x) {
      vertex_set_insert(solver->allowed_neighbors + x,
                        solver->quads[4*quad_id + i]);
    }

    if (!vertex_set_empty(delta->allowed_set_changes + i)) {
      vertex_set_union(solver->allowed_neighbors +
                       solver->quads[4*quad_id + i],
                       delta->allowed_set_changes + i);
    }

    if (!vertex_set_empty(delta->new_neighbors + i)) {
      vertex_set_subtract(solver->known_neighbors +
                          solver->quads[4*quad_id + i],
                          delta->new_neighbors + i);
    }
  }

  for (uint32_t i = 0; i < 2; i++) {
    if (delta->diagonal_mask & (1 << i)) {
      vertex_index v = solver->quads[4*quad_id + 2 + i];
      vertex_index u = solver->quads[4*quad_id + (i ^ 1)];

      vertex_set_remove(solver->known_diagonals + v, u);
      vertex_set_remove(solver->known_diagonals + u, v);
    }
  }
}

static bool check_quad_diagonal(solver2d *solver, fixed_point_delta2d *delta,
                                quad_index id,
                                vertex_index a, vertex_index b) {
  return !vertex_set_contains(solver->known_diagonals + a, b) &&
    solver->edge_map[diagonal_index(solver->num_vertices, a, b)] == NO_EDGE;
}

static bool finish_quad_diagonal(solver2d *solver, fixed_point_delta2d *delta,
                                 quad_index quad_id,
                                 vertex_index u, vertex_index v,
                                 quad_vertex_index i) {
  if (!check_quad_diagonal(solver, delta, quad_id, u, v)) {
    return false;
  }
  else {
    delta->diagonal_mask |= (1 << i);
    vertex_set_insert(solver->known_diagonals + u, v);
    vertex_set_insert(solver->known_diagonals + v, u);
  }

  return true;
}

static void finish_ordered_edge2d(solver2d *solver, fixed_point_delta2d *delta,
                                  quad_vertex_index i, quad_vertex_index j,
                                  vertex_index v, vertex_index u);

static bool finish_edge2d(solver2d *solver, fixed_point_delta2d *delta,
                          quad_vertex_index i, quad_vertex_index j,
                          vertex_index u, vertex_index v) {
  finish_ordered_edge2d(solver, delta, i, j, u, v);
  finish_ordered_edge2d(solver, delta, j, i, v, u);

  uint32_t edge_id = 2;
  if ((i == 3 && j == 1) || (i == 1 && j == 3))
    edge_id = 3;
  else if ((i == 3 && j == 2) || (i == 2 && j == 3))
    edge_id = 4;

  if ((solver->bound_mask & (1 << edge_id)) == 0) {
    solver->bound_mask |= (1 << edge_id);
    delta->new_bound_mask |= (1 << edge_id);

    edge_index e = solver->edge_map[diagonal_index(solver->num_vertices, u, v)];
    if (e == NO_EDGE) {
      if (solver->max_edge_id == solver->num_edges)
        return false;

      delta->max_edge_change++;
      e = solver->max_edge_id++;

      solver->edge_map[diagonal_index(solver->num_vertices, u, v)] = e;
      solver->edges[2*e + 0] = u;
      solver->edges[2*e + 1] = v;
      solver->edge_queue[solver->edge_queue_size++] = e;
    }

    if (solver->num_occurrences[e] == 0)
      return false;

    solver->num_occurrences[e]--;
    delta->occurred[delta->num_occurrences_changes++] = e;
  }

  return true;
}

static void finish_ordered_edge2d(solver2d *solver, fixed_point_delta2d *delta,
                                  quad_vertex_index i, quad_vertex_index j,
                                  vertex_index a, vertex_index b) {
  if (vertex_set_insert(solver->known_neighbors + a, b)) {
    vertex_set_insert(delta->new_neighbors + i, b);
    VERTEX_SET_FOREACH(solver->known_neighbors[a], v) {
      if (vertex_set_remove(solver->allowed_neighbors + v, b))
        vertex_set_insert(delta->removed_from + j, v);
    }

    vertex_set removed = solver->allowed_neighbors[a];
    vertex_set_intersect(&removed, solver->known_neighbors + b);
    vertex_set_union(delta->allowed_set_changes + i, &removed);
    vertex_set_subtract(solver->allowed_neighbors + a, &removed);
  }
}

static void forbid_ordered_edge2d(solver2d *solver, fixed_point_delta2d *delta,
                                  quad_vertex_index i, quad_vertex_index j,
                                  vertex_index a, vertex_index b);

static void forbid_edge2d(solver2d *solver, fixed_point_delta2d *delta,
                          quad_vertex_index i, quad_vertex_index j,
                          vertex_index a, vertex_index b) {
  forbid_ordered_edge2d(solver, delta, i, j, a, b);
  forbid_ordered_edge2d(solver, delta, j, i, b, a);
}

static void forbid_ordered_edge2d(solver2d *solver, fixed_point_delta2d *delta,
                                  quad_vertex_index i, quad_vertex_index j,
                                  vertex_index a, vertex_index b) {
  if (vertex_set_remove(solver->allowed_neighbors + a, b)) {
    vertex_set_insert(delta->removed_from + i, b);
  }
}
