/*
 * topological-hex, a program to compute combinatorial hexahedral meshes.
 *
 * Copyright (C) <2018> <Université catholique de Louvain (UCL), Belgique>
 *
 * List of the contributors to the development, description and complete
 * License: see LICENSE file.
 *
 * This program (topological-hex) is free software:
 * you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see COPYING file).  If not,
 * see <http://www.gnu.org/licenses/>.
 */


#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include "mesh.h"

static error_code read_vertices(mesh *mesh, FILE *file);
static error_code read_quads(mesh *mesh, FILE *file);
static error_code read_hexes(mesh *mesh, FILE *file);

void mesh_init(mesh *mesh) {
  mesh->num_vertices       = 0;
  mesh->num_boundary_quads = 0;
  mesh->num_quads          = 0;
  mesh->num_hexes          = 0;

  mesh->vertices      = NULL;
  mesh->quads         = NULL;
  mesh->hexes         = NULL;
  mesh->hex_adjacency = NULL;
  mesh->hex_facets    = NULL;
}

error_code mesh_load(mesh *mesh, const char *path) {
  error_code status = SUCCESS;

  mesh_init(mesh);

  FILE *file = fopen(path, "r");
  if (!file)
    return IO_ERROR;

  char buffer[256];
  while (fgets(buffer, sizeof(buffer), file) && status == SUCCESS) {
    char *it = buffer;
    while (*it && isspace(*it)) it++;

    const char vertices_cmd[] = "Vertices";
    const char quads_cmd[]    = "Quads";
    const char quads_cmd2[]   = "Quadrilaterals";
    const char hexes_cmd[]    = "Hexahedra";

    if (memcmp(it, vertices_cmd, sizeof(vertices_cmd)-1) == 0)
      status = read_vertices(mesh, file);
    else if (memcmp(it, quads_cmd, sizeof(quads_cmd)-1) == 0 ||
             memcmp(it, quads_cmd2, sizeof(quads_cmd2)-1) == 0)
      status = read_quads(mesh, file);
    else if (memcmp(it, hexes_cmd, sizeof(hexes_cmd)-1) == 0)
      status = read_hexes(mesh, file);
  }

  if (status == SUCCESS && !feof(file))
    status = IO_ERROR;

  if (status != SUCCESS)
    mesh_release(mesh);

  fclose(file);

  return status;
}

error_code mesh_reserve(mesh *mesh, uint32_t num_hexes) {
  uint32_t num_interior_quads = 3*num_hexes - mesh->num_boundary_quads/2;
  uint32_t num_quads = mesh->num_boundary_quads + num_interior_quads;

  vertex_index *quads;
  if (posix_memalign((void**)&quads, ALIGNMENT,
                     4*sizeof(*quads)*num_quads) != 0) {
    return OUT_OF_MEMORY;
  }

  if (mesh->quads) {
    memcpy(quads, mesh->quads, 4*mesh->num_quads*sizeof(*quads));
    free(mesh->quads);
  }

  mesh->quads = quads;

  vertex_index *hexes;
  if (posix_memalign((void**)&hexes, ALIGNMENT,
                     8*sizeof(*hexes)*num_hexes) != 0) {
    return OUT_OF_MEMORY;
  }

  if (mesh->hexes) {
    memcpy(hexes, mesh->hexes, 8*mesh->num_hexes*sizeof(*hexes));
    free(mesh->hexes);
  }

  mesh->hexes = hexes;

  quad_index *hex_facets;
  if (posix_memalign((void**)&hex_facets, ALIGNMENT,
                     6*sizeof(*hex_facets)*num_hexes) != 0) {
    return OUT_OF_MEMORY;
  }

  if (mesh->hex_facets) {
    memcpy(hex_facets, mesh->hex_facets, 6*mesh->num_hexes*sizeof(*hex_facets));
    free(mesh->hex_facets);
  }

  mesh->hex_facets = hex_facets;

  hex_index *hex_adjacency;
  if (posix_memalign((void**)&hex_adjacency, ALIGNMENT,
                     6*sizeof(*hex_adjacency)*num_hexes) != 0) {
    return OUT_OF_MEMORY;
  }

  if (mesh->hex_adjacency) {
    memcpy(hex_adjacency, mesh->hex_adjacency,
           6*mesh->num_hexes*sizeof(*hex_adjacency));
    free(mesh->hex_adjacency);
  }

  mesh->hex_adjacency = hex_adjacency;

  mesh->num_quads = num_quads;
  mesh->num_hexes = num_hexes;

  return SUCCESS;
}

void mesh_release(mesh *mesh) {
  if (mesh->hex_facets) free(mesh->hex_facets);
  if (mesh->hex_adjacency) free(mesh->hex_adjacency);
  if (mesh->hexes) free(mesh->hexes);
  if (mesh->quads) free(mesh->quads);
  if (mesh->vertices) free(mesh->vertices);
}

static error_code read_vertices(mesh *mesh, FILE *file) {
  char buffer[256];
  if (!fgets(buffer, sizeof(buffer), file))
    return IO_ERROR;

  char *end = NULL;
  mesh->num_vertices = strtoul(buffer, &end, 10);
  if (!end || (*end != '\0' && *end != '\n'))
    return PARSE_ERROR;

  if (posix_memalign((void**)&mesh->vertices, ALIGNMENT,
                     4*sizeof(*mesh->vertices)*mesh->num_vertices) != 0) {
    return OUT_OF_MEMORY;
  }

  for (uint32_t i = 0; i < mesh->num_vertices; i++) {
    if (!fgets(buffer, sizeof(buffer), file))
      return IO_ERROR;

    float data[3];
    if (sscanf(buffer, "%f %f %f 0", &data[0], &data[1], &data[2]) != 3)
      return PARSE_ERROR;

    for (int j = 0; j < 3; j++) mesh->vertices[4*i + j] = data[j];
    mesh->vertices[4*i + 3] = 0.0;
  }

  return SUCCESS;
}

static error_code read_quads(mesh *mesh, FILE *file) {
  char buffer[256];
  if (!fgets(buffer, sizeof(buffer), file))
    return IO_ERROR;

  char *end = NULL;
  mesh->num_boundary_quads = strtoul(buffer, &end, 10);
  mesh->num_quads = mesh->num_boundary_quads;
  if (!end || (*end != '\0' && *end != '\n'))
    return PARSE_ERROR;

  if (posix_memalign((void**)&mesh->quads, ALIGNMENT,
                     4*sizeof(*mesh->quads)*mesh->num_quads) != 0) {
    return OUT_OF_MEMORY;
  }

  for (uint32_t i = 0; i < mesh->num_quads; i++) {
    if (!fgets(buffer, sizeof(buffer), file))
      return IO_ERROR;

    uint32_t quad[4];
    if (sscanf(buffer, "%u %u %u %u 0",
               quad + 0, quad + 1, quad + 2, quad + 3) != 4)
      return PARSE_ERROR;

    for (quad_vertex_index j = 0; j < 4; j++)
      mesh->quads[4*i + j] = quad[j] - 1;
  }

  return SUCCESS;
}

static error_code read_hexes(mesh *mesh, FILE *file) {
  char buffer[256];
  if (!fgets(buffer, sizeof(buffer), file))
    return IO_ERROR;

  char *end = NULL;
  mesh->num_hexes = strtoul(buffer, &end, 10);
  if (!end || (*end != '\0' && *end != '\n'))
    return PARSE_ERROR;

  if (posix_memalign((void**)&mesh->hexes, ALIGNMENT,
                     8*sizeof(*mesh->hexes)*mesh->num_hexes) != 0) {
    return OUT_OF_MEMORY;
  }

  if (posix_memalign((void**)&mesh->hex_adjacency, ALIGNMENT,
                     6*sizeof(*mesh->hex_adjacency)*mesh->num_hexes) != 0) {
    return OUT_OF_MEMORY;
  }
  memset(mesh->hex_adjacency, 0,
         6*sizeof(*mesh->hex_adjacency)*mesh->num_hexes);

  if (posix_memalign((void**)&mesh->hex_facets, ALIGNMENT,
                     6*sizeof(*mesh->hex_facets)*mesh->num_hexes) != 0) {
    return OUT_OF_MEMORY;
  }
  memset(mesh->hex_facets, 0, 6*sizeof(*mesh->hex_facets)*mesh->num_hexes);

  for (uint32_t i = 0; i < mesh->num_hexes; i++) {
    if (!fgets(buffer, sizeof(buffer), file))
      return IO_ERROR;

    uint32_t hex[8];
    if (sscanf(buffer, "%u %u %u %u %u %u %u %u 0",
               hex + 0, hex + 1, hex + 2, hex + 3,
               hex + 4, hex + 5, hex + 6, hex + 7) != 8)
      return PARSE_ERROR;

    for (hex_vertex_index j = 0; j < 8; j++)
      mesh->hexes[8*i + j] = hex[j] - 1;
  }

  return SUCCESS;
}

uint32_t mesh_num_boundary_vertices(const mesh *mesh) {
  uint32_t n = 0;
  for (uint32_t i = 0; i < 4*mesh->num_boundary_quads; i++) {
    if (n < mesh->quads[i]) n = mesh->quads[i];
  }

  return n + 1;
}
