/*
 * topological-hex, a program to compute combinatorial hexahedral meshes.
 *
 * Copyright (C) <2018> <Université catholique de Louvain (UCL), Belgique>
 *
 * List of the contributors to the development, description and complete
 * License: see LICENSE file.
 *
 * This program (topological-hex) is free software:
 * you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see COPYING file).  If not,
 * see <http://www.gnu.org/licenses/>.
 */


#include "mesh2d.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

static error_code read_vertices(mesh2d *mesh, FILE *file);
static error_code read_edges(mesh2d *mesh, FILE *file);
static error_code read_quads(mesh2d *mesh, FILE *file);


void mesh2d_init(mesh2d *mesh) {
  mesh->num_vertices       = 0;
  mesh->num_boundary_edges = 0;
  mesh->num_edges          = 0;
  mesh->num_quads          = 0;

  mesh->vertices   = NULL;
  mesh->edges      = NULL;
  mesh->quads      = NULL;
  mesh->quad_edges = NULL;
}

error_code mesh2d_load(mesh2d *mesh, const char *path) {
  error_code status = SUCCESS;

  mesh2d_init(mesh);

  FILE *file = fopen(path, "r");
  if (!file)
    return IO_ERROR;

  char buffer[256];
  while (fgets(buffer, sizeof(buffer), file) && status == SUCCESS) {
    const char vertices_cmd[] = "Vertices";
    const char edges_cmd[]    = "Edges";
    const char quads_cmd[]    = "Quads";
    const char quads_cmd2[]   = "Quadrilaterals";

    if (memcmp(buffer, vertices_cmd, sizeof(vertices_cmd)-1) == 0)
      status = read_vertices(mesh, file);
    else if (memcmp(buffer, edges_cmd, sizeof(edges_cmd)-1) == 0)
      status = read_edges(mesh, file);
    else if (memcmp(buffer, quads_cmd, sizeof(quads_cmd)-1) == 0 ||
             memcmp(buffer, quads_cmd2, sizeof(quads_cmd2)-1) == 0)
      status = read_quads(mesh, file);
  }

  if (status == SUCCESS && !feof(file))
    status = IO_ERROR;

  if (status != SUCCESS)
    mesh2d_release(mesh);

  fclose(file);

  return status;
}

error_code mesh2d_reserve(mesh2d *mesh, uint32_t num_quads) {
  uint32_t num_interior_edges = 2*num_quads - mesh->num_boundary_edges/2;

  uint32_t num_edges = mesh->num_boundary_edges + num_interior_edges;

  vertex_index *edges;
  if (posix_memalign((void**)&edges, ALIGNMENT,
                     2 * sizeof(*edges) * num_edges) != 0) {
    return OUT_OF_MEMORY;
  }

  if (mesh->edges) {
    memcpy(edges, mesh->edges, 2 * sizeof(*edges) * mesh->num_edges);
    free(mesh->edges);
  }

  mesh->edges = edges;
  mesh->num_edges = num_edges;

  vertex_index *quads;
  if (posix_memalign((void**)&quads, ALIGNMENT,
                     4 * sizeof(*quads) * num_quads) != 0) {
    return OUT_OF_MEMORY;
  }

  if (mesh->quads) {
    memcpy(quads, mesh->quads, 4 * sizeof(*quads) * num_quads);
    free(mesh->quads);
  }

  mesh->quads = quads;

  edge_index *quad_edges;
  if (posix_memalign((void**)&quad_edges, ALIGNMENT,
                     4 * sizeof(*quad_edges) * num_quads) != 0) {
    return OUT_OF_MEMORY;
  }

  if (mesh->quad_edges) {
    memcpy(quad_edges, mesh->quad_edges, 4 * sizeof(*quad_edges) * num_quads);
    free(mesh->quad_edges);
  }

  mesh->quad_edges = quad_edges;

  mesh->num_quads = num_quads;

  return SUCCESS;
}

void mesh2d_release(mesh2d *mesh) {
  free(mesh->vertices);
  free(mesh->edges);
  free(mesh->quads);
  free(mesh->quad_edges);
}

uint32_t mesh2d_num_boundary_vertices(const mesh2d *mesh) {
  uint32_t n = 0;
  for (uint32_t i = 0; i < 2*mesh->num_boundary_edges; i++) {
    if (n < mesh->edges[i]) n = mesh->edges[i];
  }

  return n + 1;
}

static error_code read_vertices(mesh2d *mesh, FILE *file) {
  char buffer[256];
  if (!fgets(buffer, sizeof(buffer), file))
    return IO_ERROR;

  char *end = NULL;
  mesh->num_vertices = strtoul(buffer, &end, 10);
  if (!end || (*end != '\0' && *end != '\n'))
    return PARSE_ERROR;

  if (posix_memalign((void**)&mesh->vertices, ALIGNMENT,
                     4*sizeof(*mesh->vertices)*mesh->num_vertices) != 0) {
    return OUT_OF_MEMORY;
  }

  for (uint32_t i = 0; i < mesh->num_vertices; i++) {
    if (!fgets(buffer, sizeof(buffer), file))
      return IO_ERROR;

    float data[3];
    if (sscanf(buffer, "%f %f %f 0", &data[0], &data[1], &data[2]) != 3)
      return PARSE_ERROR;

    for (int j = 0; j < 3; j++) mesh->vertices[4*i + j] = data[j];
    mesh->vertices[4*i + 3] = 0.0;
  }

  return SUCCESS;
}

static error_code read_edges(mesh2d *mesh, FILE *file) {
  char buffer[256];
  if (!fgets(buffer, sizeof(buffer), file))
    return IO_ERROR;

  char *end = NULL;
  mesh->num_boundary_edges = strtoul(buffer, &end, 10);
  if (!end || (*end != '\0' && *end != '\n'))
    return PARSE_ERROR;

  mesh->num_edges = mesh->num_boundary_edges;

  if (posix_memalign((void**)&mesh->edges, ALIGNMENT,
                     4*sizeof(*mesh->edges)*mesh->num_edges) != 0) {
    return OUT_OF_MEMORY;
  }

  for (uint32_t i = 0; i < mesh->num_edges; i++) {
    if (!fgets(buffer, sizeof(buffer), file))
      return IO_ERROR;

    uint32_t edge[2];
    if (sscanf(buffer, "%u %u 0", edge + 0, edge + 1) != 2)
      return PARSE_ERROR;

    for (uint32_t j = 0; j < 2; j++)
      mesh->edges[2*i + j] = edge[j] - 1;
  }

  return SUCCESS;
}

static error_code read_quads(mesh2d *mesh, FILE *file) {
  char buffer[256];
  if (!fgets(buffer, sizeof(buffer), file))
    return IO_ERROR;

  char *end = NULL;
  mesh->num_quads = strtoul(buffer, &end, 10);
  if (!end || (*end != '\0' && *end != '\n'))
    return PARSE_ERROR;

  if (posix_memalign((void**)&mesh->quads, ALIGNMENT,
                     4*sizeof(*mesh->quads)*mesh->num_quads) != 0) {
    return OUT_OF_MEMORY;
  }

  if (posix_memalign((void**)&mesh->quad_edges, ALIGNMENT,
                     4*sizeof(*mesh->quad_edges)*mesh->num_quads) != 0) {
    return OUT_OF_MEMORY;
  }

  for (uint32_t i = 0; i < mesh->num_quads; i++) {
    if (!fgets(buffer, sizeof(buffer), file))
      return IO_ERROR;

    uint32_t quad[4];
    if (sscanf(buffer, "%u %u %u %u 0",
               quad + 0, quad + 1, quad + 2, quad + 3) != 4)
      return PARSE_ERROR;

    for (quad_vertex_index j = 0; j < 4; j++)
      mesh->quads[4*i + j] = quad[j] - 1;
  }

  return SUCCESS;
}
