/*
 * topological-hex, a program to compute combinatorial hexahedral meshes.
 *
 * Copyright (C) <2018> <Université catholique de Louvain (UCL), Belgique>
 *
 * List of the contributors to the development, description and complete
 * License: see LICENSE file.
 *
 * This program (topological-hex) is free software:
 * you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see COPYING file).  If not,
 * see <http://www.gnu.org/licenses/>.
 */


#ifndef MESH2D_H_
#define MESH2D_H_

#include <stdint.h>

#include "types.h"
#include "utils.h"

/**
 * @defgroup mesh2d 2D Mesh
 *
 * Data structure and basic operations to manipulate a 2D mesh.
 */

/**
 * @addtogroup mesh2d
 * @{
 */

typedef struct mesh2d {
  uint32_t num_vertices;

  /**
   * Array containing the coordinates of each vertex. 4 values per element, the
   * fourth of which being unused.
   */
  ALIGNED double *vertices;

  /**
   * Number of edges on the boundary.
   */
  uint32_t num_boundary_edges;

  /**
   * Total number of edges. This includes @ref mesh::num_boundary_quads.
   */
  uint32_t num_edges;

  uint32_t num_quads;

  /**
   * Indices of the vertices of each edge.
   */
  ALIGNED vertex_index *edges;

  /**
   * Indices of the vertices of each quad.
   */
  ALIGNED vertex_index *quads;

  /**
   * Indices of the edges of each quad.
   */
  ALIGNED edge_index *quad_edges;
} mesh2d;

/**
 * Initializes a mesh to be empty.
 *
 * Calling this function is not necessary when using @ref mesh2d_load.
 */
void mesh2d_init(mesh2d *mesh);

/**
 * Loads a mesh from a file.
 *
 * The contents of the mesh data structure do not need to be initialized when
 * this function is called.
 */
error_code mesh2d_load(mesh2d *mesh, const char *path);

/**
 * Reserves space for a certain number of quads within the mesh.
 *
 * This function can only be used to increase the total number of quads. It
 * also does not remove existing data from the mesh.
 *
 * This function should be called on an initialized mesh.
 */
error_code mesh2d_reserve(mesh2d *mesh, uint32_t num_quads);

/**
 * Releases all memory used to store a mesh.
 */
void mesh2d_release(mesh2d *mesh);

/**
 * Returns the number of vertices on the boundary of the mesh.
 *
 * This operation only works if all vertices on this boundary are numbered from
 * 0 to n without gaps.
 */
uint32_t mesh2d_num_boundary_vertices(const mesh2d *mesh);

/**
 * @}
 */
#endif
