/*
 * topological-hex, a program to compute combinatorial hexahedral meshes.
 *
 * Copyright (C) <2018> <Université catholique de Louvain (UCL), Belgique>
 *
 * List of the contributors to the development, description and complete
 * License: see LICENSE file.
 *
 * This program (topological-hex) is free software:
 * you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see COPYING file).  If not,
 * see <http://www.gnu.org/licenses/>.
 */


#ifndef SOLVER_H_
#define SOLVER_H_

#include <stdint.h>
#include <stddef.h>

#include "types.h"
#include "vertex_set.h"
#include "mesh.h"

/**
 * @defgroup solver Solver
 *
 * Data structures and operations used to find all-hexahedral meshes
 * topologically, ignoring the coordinates of the vertices.
 */

/**
 * @addtogroup solver
 * @{
 */

struct solver;

/**
 * Function called when a solution is found.
 *
 * @param num_hexes Number of hexahedra in the solution
 * @param data The user_data pointer passed to @ref solver_run.
 */
typedef void (*solver_callback)(const struct solver *solver,
                                uint32_t num_hexes, void *user_data);

typedef struct solver {
  uint32_t num_vertices;
  uint32_t num_quads;
  uint32_t num_hexes;
  size_t num_branches;

  uint32_t max_quad_id, max_vertex_id;

  hex_index hex_id;

  ALIGNED vertex_index *quads;
  ALIGNED vertex_index *hexes;
  ALIGNED hex_index *hex_adjacency;
  ALIGNED quad_index *hex_facets;

  uint32_t quad_queue_size;
  uint32_t quad_queue_offset;
  ALIGNED quad_index *quad_queue;
  ALIGNED uint32_t *num_occurrences;

  ALIGNED quad_descriptor *quad_descriptors;

  ALIGNED vertex_set *allowed_neighbors;
  ALIGNED vertex_set *known_neighbors;
  ALIGNED vertex_set *known_diagonals;

  ALIGNED vertex_set domains[4];
  uint32_t bound_mask;

  ALIGNED vertex_index *no_goods;
  size_t no_goods_size, no_goods_capacity;

  ALIGNED vertex_index *boundary_group;
  size_t num_symmetries;
  uint32_t num_boundary_vertices;

  solver_callback cb;
  void *user_data;
} solver;

/**
 * Allocates data to find hexahedral meshes.
 *
 * @param mesh The mesh to manipulate. Pointers within this data structure will
 *   be used while the solver tries to fill them, and therefore the mesh must be
 *   released after the solver.
 * @param num_vertices Maximum number of vertices in a solution. This value
 *   includes both boundary and internal vertices.
 */
error_code solver_init(solver *solver, mesh *mesh, uint32_t num_vertices);

/**
 * Releases all memory used by a solver.
 */
void solver_release(solver *solver);

/**
 * Initializes a copy of a solver.
 *
 * This copy can be used independently of the other solver. This includes using
 * both solvers in parallel.
 *
 * Such a copy must be released with @ref solver_release_copy, not @ref
 * solver_release.
 */
error_code solver_init_copy(struct solver *solver, const struct solver *other);

/**
 * Releases all memory used by a solver initialized using @ref solver_init_copy.
 */
void solver_release_copy(solver *solver);

void solver_copy(struct solver *solver, const struct solver *other);

/**
 * Special indicaters used to encode the tree of no-goods in a linear structure.
 */
typedef enum no_goods_instruction {
  /** Followed by a vertex assignment */
  NO_GOODS_VERTEX_0,
  NO_GOODS_VERTEX_1,
  NO_GOODS_VERTEX_2,
  NO_GOODS_VERTEX_3,

  /** Followed by the children of the current node */
  NO_GOODS_DOWN,

  /** Followed by the four vertices that make up the face of the next
   *  hexahedron */
  NO_GOODS_FACE,
} no_goods_instruction;

/**
 * Stores the changes performed by @ref solver_start_branch so they can be
 * cancelled by @ref solver_undo_no_goods.
 */
typedef struct no_goods_delta {
  size_t size;
  hex_index old_hex_index;
  uint8_t old_bound_mask;
} no_goods_delta;

/**
 * Configures the set of permutations that preserve the face lattice of the
 * input.
 *
 * This can save time by skipping subtrees symmetrical to already explored
 * portions of the search space. The requirement for this operation to be valid
 * is that permutation.
 *
 * @param symmetries The permutations to use as symmetries. Each permutation is
 *   encoded as an array `a` where `a[i] = j` means that the permutation
 *   replaces i by j.
 */
error_code solver_set_symmetries(solver *solver,
                                 size_t num_symmetries,
                                 uint32_t num_boundary_vertices,
                                 const vertex_index *symmetries);

/**
 * Called before each branching node to maintain the state of no-goods.
 *
 * This is used to break symmetries: if a node is a subtree of a no-good (up to
 * symmetry), it can be pruned.
 *
 * When backtracking, this operation can be cancelled along with subsequent
 * calls to @ref solver_is_no_good.
 *
 * @param [out] delta Changes performed by this function, so that they can
 *   be cancelled.
 */
error_code solver_start_branch(solver *solver, no_goods_delta *delta);

/**
 * Indicates that binding @p i to @p v from the current node produces a subtree
 * that has already been explored.
 *
 * Changes are added to the existing delta structure so that a single function
 * call can undo both new and existing change.
 *
 * @see solver_start_branch
 */
error_code solver_is_no_good(solver *solver,
                             quad_vertex_index i, vertex_index v,
                             no_goods_delta *delta);

/**
 * Cancels calls to @ref solver_start_branch and @ref solver_is_no_good.
 */
void solver_undo_no_goods(solver *solver, const no_goods_delta *delta);

/**
 * Prints debugging information about a solver.
 *
 * @param num_hexes Number of completed hexahedra
 * @param header C-style string printed to give context about how this state was
 *   obtained.
 */
void solver_print(solver *solver, uint32_t num_hexes, const char *header);

typedef struct solver_pool {
  void *data;
  uint32_t element_size, capacity;
} solver_pool;

error_code solver_pool_alloc(solver_pool *pool,
                             uint32_t num_vertices,
                             uint32_t num_quads,
                             uint32_t num_hexes,
                             uint32_t num_symmetries,
                             uint32_t num_boundary_vertices,
                             uint32_t n);
error_code solver_pool_init(solver_pool *pool, const struct solver *template,
                            uint32_t n);
void solver_pool_release(solver_pool *pool);

error_code solver_pool_realloc(solver_pool *pool, uint32_t n);

static inline
solver *solver_pool_get(solver_pool *pool, size_t i) {
  return (solver*)((char*)pool->data + pool->element_size*i);
}

/**
 * Loads the set of symmetries from a file.
 *
 * @param [out] vertices Array of permutations in the symmetry group. Should be
 *   freed by the caller.
 * @param [out] num_boundary_vertices Number of vertices per permutation
 * @param [out] num_symmetries Number of permutations in the symmetries group
 */
error_code read_symmetries(const char *path, vertex_index **vertices,
                           vertex_index *num_boundary_vertices,
                           size_t *num_symmetries);

/**
 * @}
 */

#endif
