/*
 * topological-hex, a program to compute combinatorial hexahedral meshes.
 *
 * Copyright (C) <2018> <Université catholique de Louvain (UCL), Belgique>
 *
 * List of the contributors to the development, description and complete
 * License: see LICENSE file.
 *
 * This program (topological-hex) is free software:
 * you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see COPYING file).  If not,
 * see <http://www.gnu.org/licenses/>.
 */


#include <stdlib.h>
#include <string.h>

#include "solver2d.h"

error_code solver2d_init(solver2d *solver, mesh2d *mesh, uint32_t num_vertices,
                         uint32_t num_valences, uint32_t *valences) {
  solver->num_vertices = num_vertices;
  solver->num_boundary_vertices = mesh2d_num_boundary_vertices(mesh);
  /* solver->num_known_vertices = mesh->num_vertices; */
  solver->num_known_vertices = solver->num_boundary_vertices;
  solver->num_edges = mesh->num_edges;
  solver->num_quads = mesh->num_quads;

  solver->max_edge_id = mesh->num_boundary_edges;

  solver->vertices = mesh->vertices;
  solver->edges = mesh->edges;
  solver->quads = mesh->quads;
  solver->quad_edges = mesh->quad_edges;

  solver->edge_queue_size = mesh->num_boundary_edges;
  solver->edge_queue_offset = 0;
  if (posix_memalign((void**)&solver->edge_queue, ALIGNMENT,
                     sizeof(*solver->edge_queue) * mesh->num_edges) != 0) {
    goto fail_alloc_queue;
  }

  for (uint32_t i = 0; i < mesh->num_boundary_edges; i++)
    solver->edge_queue[i] = i;

  if (posix_memalign((void**)&solver->num_occurrences, ALIGNMENT,
                     sizeof(*solver->num_occurrences) * mesh->num_edges) != 0) {
    goto fail_alloc_num_occurrences;
  }

  for (uint32_t i = 0; i < mesh->num_boundary_edges; i++)
    solver->num_occurrences[i] = 1;
  for (uint32_t i = mesh->num_boundary_edges; i < mesh->num_edges; i++)
    solver->num_occurrences[i] = 2;

  if (posix_memalign(
        (void**)&solver->edge_map, ALIGNMENT,
        sizeof(*solver->edge_map) * num_vertices * num_vertices) != 0) {
    goto fail_alloc_edge_map;
  }

  for (uint32_t i = 0; i < num_vertices * num_vertices; i++) {
    solver->edge_map[i] = NO_EDGE;
  }

  for (edge_index i = 0; i < mesh->num_boundary_edges; i++) {
    vertex_index a = mesh->edges[2*i + 0];
    vertex_index b = mesh->edges[2*i + 1];
    solver->edge_map[diagonal_index(num_vertices, a, b)] = i;
  }

  if (posix_memalign(
        (void**)&solver->valences, ALIGNMENT,
        sizeof(*solver->valences)*num_vertices) != 0) {
    goto fail_alloc_valences;
  }

  if (posix_memalign(
        (void**)&solver->original_valences, ALIGNMENT,
        sizeof(*solver->valences)*num_vertices) != 0) {
    goto fail_alloc_original_valences;
  }

  memcpy(solver->valences, valences, num_valences * sizeof(*valences));
  for (vertex_index i = num_valences; i < num_vertices; i++)
    solver->valences[i] = 4;

  memcpy(solver->original_valences, solver->valences,
         num_vertices * sizeof(*valences));

  for (uint32_t i = 0;
       i < sizeof(solver->for_valence)/sizeof(*solver->for_valence); i++) {
    solver->for_valence[i].used = false;
    solver->for_valence[i].max_vertex_id = solver->num_boundary_vertices;
    solver->for_valence[i].limit = 0;
  }

  for (vertex_index v = mesh2d_num_boundary_vertices(mesh);
       v < num_valences; v++) {
    uint32_t i = solver->valences[v];

    solver->for_valence[i].used = true;

    for (uint32_t j = i + 1;
         j < sizeof(solver->for_valence)/sizeof(*solver->for_valence); j++) {
      if (solver->for_valence[j].max_vertex_id < v + 1u)
        solver->for_valence[j].max_vertex_id = v + 1u;
    }

    if (solver->for_valence[i].limit < v +1u)
      solver->for_valence[i].limit = v + 1u;
  }

  solver->for_valence[4].used = true;
  solver->for_valence[4].max_vertex_id = num_valences;
  solver->for_valence[4].limit = num_vertices;

  if (posix_memalign(
        (void**)&solver->allowed_neighbors, ALIGNMENT,
        sizeof(*solver->allowed_neighbors)*num_vertices) != 0) {
    goto fail_alloc_allowed_neighbors;
  }

  if (posix_memalign(
        (void**)&solver->known_neighbors, ALIGNMENT,
        sizeof(*solver->known_neighbors)*num_vertices) != 0) {
    goto fail_alloc_known_neighbors;
  }

  if (posix_memalign(
        (void**)&solver->known_diagonals, ALIGNMENT,
        sizeof(*solver->known_diagonals)*num_vertices) != 0) {
    goto fail_alloc_known_diagonals;
  }

  for (vertex_index i = 0; i < num_vertices; i++) {
    vertex_set_init_up_to(solver->allowed_neighbors + i, num_vertices);
    vertex_set_remove(solver->allowed_neighbors + i, i);
  }

  for (vertex_index i = 0; i < num_vertices; i++)
    vertex_set_init_empty(solver->known_neighbors + i);

  for (vertex_index i = 0; i < num_vertices; i++)
    vertex_set_init_empty(solver->known_diagonals + i);

  for (edge_index i = 0; i < mesh->num_boundary_edges; i++) {
    vertex_index a = mesh->edges[2*i + 0];
    vertex_index b = mesh->edges[2*i + 1];
    vertex_set_insert(solver->known_neighbors + a, b);
    vertex_set_insert(solver->known_neighbors + b, a);
  }

  for (vertex_index v = 0; v < num_vertices; v++) {
    VERTEX_SET_FOREACH(solver->known_neighbors[v], u) {
      vertex_set_subtract(solver->allowed_neighbors + v,
                          solver->known_neighbors + u);
    }
  }

  return SUCCESS;

  free(solver->known_diagonals);
fail_alloc_known_diagonals:
  free(solver->known_neighbors);
fail_alloc_known_neighbors:
  free(solver->allowed_neighbors);
fail_alloc_allowed_neighbors:
  free(solver->original_valences);
fail_alloc_original_valences:
  free(solver->valences);
fail_alloc_valences:
  free(solver->edge_map);
fail_alloc_edge_map:
  free(solver->num_occurrences);
fail_alloc_num_occurrences:
  free(solver->edge_queue);
fail_alloc_queue:
  return OUT_OF_MEMORY;
}

void solver2d_release(solver2d *solver) {
  free(solver->known_diagonals);
  free(solver->known_neighbors);
  free(solver->allowed_neighbors);
  free(solver->original_valences);
  free(solver->valences);
  free(solver->edge_map);
  free(solver->num_occurrences);
  free(solver->edge_queue);
}
