/*
 * topological-hex, a program to compute combinatorial hexahedral meshes.
 *
 * Copyright (C) <2018> <Université catholique de Louvain (UCL), Belgique>
 *
 * List of the contributors to the development, description and complete
 * License: see LICENSE file.
 *
 * This program (topological-hex) is free software:
 * you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see COPYING file).  If not,
 * see <http://www.gnu.org/licenses/>.
 */


#ifndef SOLVER_H_
#define SOLVER_H_

#include <stdint.h>
#include <stddef.h>

#include "types.h"
#include "vertex_set.h"
#include "mesh2d.h"

/**
 * @defgroup solver2d Solver 2D
 *
 * Data structures and operations used to find all-quad meshes
 * topologically, ignoring the coordinates of the vertices.
 */

/**
 * @addtogroup solver2d
 * @{
 */

struct solver2d;

/**
 * Function called when a solution is found.
 *
 * @param num_quads Number of quads in the solution
 * @param data The user_data pointer passed to @ref solver2d_run.
 */
typedef void (*solver2d_callback)(const struct solver2d *solver,
                                  uint32_t num_quads, void *user_data);

typedef struct solver2d {
  uint32_t num_vertices;
  uint32_t num_known_vertices;
  uint32_t num_boundary_vertices;
  uint32_t num_edges;
  uint32_t num_quads;
  size_t num_branches;

  uint32_t max_edge_id;

  struct {
    bool used;
    uint32_t max_vertex_id, limit;
  } for_valence[8];

  ALIGNED double *vertices;
  ALIGNED vertex_index *edges;
  ALIGNED vertex_index *quads;
  ALIGNED edge_index *quad_edges;

  uint32_t edge_queue_size;
  uint32_t edge_queue_offset;
  ALIGNED edge_index *edge_queue;
  ALIGNED uint32_t *num_occurrences;

  ALIGNED edge_index *edge_map;

  ALIGNED uint32_t *original_valences;
  ALIGNED uint32_t *valences;
  ALIGNED vertex_set *allowed_neighbors;
  ALIGNED vertex_set *known_neighbors;
  ALIGNED vertex_set *known_diagonals;

  ALIGNED vertex_set domains[2];
  uint32_t bound_mask;

  solver2d_callback cb;
  void *user_data;
} solver2d;

/**
 * Allocates data to find quadrangular meshes.
 *
 * @param mesh The mesh to manipulate. Pointers within this data structure will
 *   be used while the solver tries to fill them, and therefore the mesh must be
 *   released after the solver.
 * @param num_vertices Maximum number of vertices in a solution. This value
 *   includes both boundary and internal vertices.
 * @param num_valences Number of vertices for which a valence constraint is
 *   known.
 * @param valences Array of size @p num_valences containing the valence for each
 *   non-regular vertex. All other vertices are assumed to be regular. This
 *   array should be sorted in increasing order.
 */
error_code solver2d_init(solver2d *solver, mesh2d *mesh, uint32_t num_vertices,
                         uint32_t num_valences, uint32_t *valences);

/**
 * Releases all memory used by a solver.
 */
void solver2d_release(solver2d *solver);

/**
 * @}
 */

#endif
