/*
 * topological-hex, a program to compute combinatorial hexahedral meshes.
 *
 * Copyright (C) <2018> <Université catholique de Louvain (UCL), Belgique>
 *
 * List of the contributors to the development, description and complete
 * License: see LICENSE file.
 *
 * This program (topological-hex) is free software:
 * you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see COPYING file).  If not,
 * see <http://www.gnu.org/licenses/>.
 */


#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <omp.h>

#include "parallel_dfs.h"

static void print_usage(const char *program_name) {
  fprintf(stderr, "Usage: %s SERVER [CHUNK_SIZE]\n", program_name);
  fprintf(stderr, "       %s SERVER INPUT H V [CHUNK_SIZE [NUM_SUBDIVISIONS "
          " [SYMMETRIES]]]\n", program_name);
  exit(1);
}

static void on_print(const char *message, void *data) {
  printf("%s", message);
}

int main(int argc, char **argv) {
  if (argc == 2 || argc == 3) {
    uint32_t chunk_size;
    if (argc == 3)
      chunk_size = atoi(argv[2]);
    else
      chunk_size = omp_get_max_threads() * 8;

    if (solver_run_distributed_client(argv[1], on_print, NULL,
                                      chunk_size) != SUCCESS) {
      fprintf(stderr, "Failed to run client connected to server %s\n", argv[1]);
      return 1;
    }

    return 0;
  }
  else {
    if (argc < 5 || argc > 8)
      print_usage(argv[0]);

    const char *server_name = argv[1];
    const char *input_name = argv[2];

    uint32_t num_hexes = atoi(argv[3]);
    uint32_t num_internal_vertices = atoi(argv[4]);

    uint32_t chunk_size;
    if (argc >= 6)
      chunk_size = atoi(argv[5]);
    else
      chunk_size = omp_get_max_threads() * 8;

    uint32_t num_subproblems;
    if (argc >= 7)
      num_subproblems = atoi(argv[6]);
    else
      num_subproblems = 512 * 1024;

    int status = 0;

    mesh mesh;
    if (mesh_load(&mesh, input_name) != SUCCESS) {
      status = 1;
      fprintf(stderr, "Failed to load mesh from file: %s\n", argv[0]);
      goto fail_mesh_load;
    }

    if (mesh_reserve(&mesh, num_hexes) != SUCCESS) {
      status = 1;
      fprintf(stderr, "Failed to reserve memory for hexes\n");
      goto fail_mesh_reserve;
    }

    solver solver;
    if (solver_init(&solver, &mesh,
                    mesh.num_vertices + num_internal_vertices) != SUCCESS) {
      status = 1;
      fprintf(stderr, "Failed to allocate memory for the solver\n");
      goto fail_solver_init;
    }

    if (argc >= 8) {
      vertex_index *symmetries;
      vertex_index num_boundary_vertices;
      size_t num_symmetries;

      if (read_symmetries(argv[7], &symmetries, &num_boundary_vertices,
                          &num_symmetries) == SUCCESS) {
        solver_set_symmetries(&solver, num_symmetries, num_boundary_vertices,
                              symmetries);
      }
      else
        fprintf(stderr, "Failed to load symmetries from file: `%s'\n", argv[4]);
    }

    if (solver_run_distributed(
          &solver, server_name,
          on_print, NULL, chunk_size, num_subproblems) != SUCCESS) {
      status = 1;
      fprintf(stderr, "Failed to run client connect to server %s\n",
              server_name);
      goto fail_run_client;
    }

  fail_run_client:
    solver_release(&solver);
  fail_solver_init:
  fail_mesh_reserve:
    mesh_release(&mesh);
  fail_mesh_load:
    return status;
  }
}
