/*
 * topological-hex, a program to compute combinatorial hexahedral meshes.
 *
 * Copyright (C) <2018> <Université catholique de Louvain (UCL), Belgique>
 *
 * List of the contributors to the development, description and complete
 * License: see LICENSE file.
 *
 * This program (topological-hex) is free software:
 * you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see COPYING file).  If not,
 * see <http://www.gnu.org/licenses/>.
 */


#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "server.h"

typedef struct solution_data {
  FILE *file;
  pthread_mutex_t *lock;
  bool wrote_progress;
  double start_time;
} solution_data;

static
void on_solution(
  const char *user_name,
  const vertex_index *vertices, hex_index num_hexes,
  void *data);

static void on_debug_log(const char *message, void *data);
static void on_log(const char *message, void *data);
static void on_listen(void *data);
static void on_progress(uint32_t i, uint32_t n, void *data);
static void on_debug_progress(uint32_t i, uint32_t n, void *data);

static double get_time(void) {
#ifdef SET_GLIBC_LINK_VERSIONS_HEADER
  return time(NULL);
#else
  struct timespec time;
  clock_gettime(CLOCK_MONOTONIC, &time);
  return time.tv_sec + 1e-9 * time.tv_nsec;
#endif
}

int status;

static void parse_args(int *argc, char **argv,
                       const char **output_path,
                       const char **backup_path,
                       bool *try_restore,
                       bool *debug);

int main(int argc, char **argv) {
  int status = 0;

  const char *output_path, *backup_path;
  bool try_restore, debug;
  parse_args(&argc, argv, &output_path, &backup_path, &try_restore, &debug);

  server_backup_info backup = {backup_path, try_restore};

  FILE *output_file;
  if (output_path)
    output_file = fopen(output_path, "w");
  else
    output_file = stdout;

  if (!output_file) {
    status = 1;
    goto fail_open_file;
  }

  pthread_mutex_t lock;
  if (pthread_mutex_init(&lock, NULL) != 0) {
    status = 1;

    fprintf(stderr, "Failed to create lock!\n");
    goto fail_create_lock;
  }

  solution_data cb_data = {output_file, &lock, false, get_time()};

  if (server_run(on_solution, &cb_data,
                 debug ? on_debug_log : on_log, &lock,
                 on_listen, NULL,
                 debug ? on_debug_progress : on_progress, &cb_data,
                 backup_path ? &backup : NULL) != SUCCESS) {
    status = 1;
    fprintf(stderr, "Failed to run server!\n");
  }

  pthread_mutex_destroy(&lock);
fail_create_lock:
  if (output_path) fclose(output_file);
fail_open_file:
  return status;
}

static bool hex_compatibility(const vertex_index *a, const vertex_index *b);

static
void on_solution(const char *user_name,
                 const vertex_index *vertices, hex_index num_hexes,
                 void *data) {
  solution_data *cb_data = data;
  pthread_mutex_lock(cb_data->lock);

  fprintf(cb_data->file, "SOLUTION FOUND BY %s!\n", user_name);
  for (hex_index i = 0; i < num_hexes; i++) {
    fprintf(cb_data->file, "  %u) {", i);
    for (hex_vertex_index j = 0; j < 8; j++) {
      fprintf(cb_data->file, "%u", vertices[8*i+j]);
      if (j + 1 != 8) fprintf(cb_data->file, ", ");
    }
    fprintf(cb_data->file, "}\n");

    for (hex_index j = i+1; j < num_hexes; j++) {
      if (!hex_compatibility(vertices + 8*i, vertices + 8*j)) {
        fprintf(cb_data->file, "  ERROR: Hex #%u and #%u are incompatible!\n",
                i, j);
      }
    }
  }
  fflush(cb_data->file);

  pthread_mutex_unlock(cb_data->lock);
}

static
void on_log(const char *message, void *data) {}

static
void on_debug_log(const char *message, void *data) {
  pthread_mutex_t *lock = data;
  pthread_mutex_lock(lock);
  printf("%s\n", message);
  pthread_mutex_unlock(lock);
}

static void on_listen(void *data) {
}

#define BAR_WIDTH 50

static void on_debug_progress(uint32_t i, uint32_t n, void *data) {}

static void on_progress(uint32_t i, uint32_t n, void *data) {
  solution_data *args = data;

  FILE *out = stdout;

  pthread_mutex_lock(args->lock);
  if (!args->wrote_progress)
    fprintf(out, "\n");

  char bar[BAR_WIDTH + 1];
  uint64_t ticks = (BAR_WIDTH * (uint64_t)i) / n;
  memset(bar, '=', ticks);
  if (i != 0 && i != n) {
    bar[ticks - 1] = '>';
    memset(bar + ticks, ' ', BAR_WIDTH - ticks);
  }
  else if (i == 0)
    memset(bar + ticks, ' ', BAR_WIDTH - ticks);
  bar[BAR_WIDTH] = 0;

  fprintf(out, "\r\033[K[%s] ", bar);

  if (i == n)
    fprintf(out, "DONE!");
  else if (i != 0) {
    double elapsed = get_time() - args->start_time;
    uint64_t eta = (n - i) * (elapsed / i);

    uint64_t days = eta / (24 * 3600);
    eta -= 24 * 3600 * days;
    uint64_t hours = eta / 3600;
    eta -= 3600 * hours;
    uint64_t minutes = eta / 60;
    uint64_t seconds = eta - 60 * minutes;

    if (days != 0)
      fprintf(out, "ETA: %zud%02zu:%02zu:%02zu", days, hours, minutes, seconds);
    else if (hours != 0)
      fprintf(out, "ETA: %02zu:%02zu:%02zu", hours, minutes, seconds);
    else if (minutes != 0)
      fprintf(out, "ETA: %02zu:%02zu", minutes, seconds);
    else
      fprintf(out, "ETA: %zus", seconds);
  }

  args->wrote_progress = true;

  fflush(out);
  pthread_mutex_unlock(args->lock);
}

static void print_usage(const char *program_name, int status) {
  printf("Usage: %s [OPTIONS] OUTPUT_PATH\n", program_name);
  printf("  -h, --help      Display this message\n");
  printf("  --restore       Try to restore progress from the backup\n");
  printf("  --debug         Logs messages received from clients\n");
  printf("  --backup DIR    Store all progress made by the server to this "
         " directory\n");
  exit(status);
}

static void parse_args(int *argc, char **argv,
                       const char **output_path,
                       const char **backup_path,
                       bool *try_restore, bool *debug) {
  *output_path = *backup_path = NULL;
  *try_restore = false;
  *debug = false;

  for (int i = 1; i < *argc; ) {
    if (strcmp(argv[i], "--restore") == 0) {
      *try_restore = true;
      memmove(argv + i, argv + i + 1, ((*argc) - i - 1) * sizeof(*argv));
      (*argc)--;
    }
    if (strcmp(argv[i], "--debug") == 0) {
      *debug = true;
      memmove(argv + i, argv + i + 1, ((*argc) - i - 1) * sizeof(*argv));
      (*argc)--;
    }
    else if (strcmp(argv[i], "--backup") == 0) {
      if (i + 1 == *argc) print_usage(argv[0], 1);

      *backup_path = argv[i + 1];

      memmove(argv + i, argv + i + 2, ((*argc) - i - 2) * sizeof(*argv));
      (*argc) -= 2;
    }
    else if (strcmp(argv[i], "--help") == 0) {
      print_usage(argv[0], 0);
    }
    else
      i++;
  }

  if (*argc >= 2)
    *output_path = argv[1];
}


static
uint32_t shared_vertices(const vertex_index *a, const vertex_index *b,
                         hex_vertex_index *id_a, hex_vertex_index *id_b);

static
bool is_edge(hex_vertex_index a, hex_vertex_index b);

static
bool is_face(hex_vertex_index a, hex_vertex_index b,
             hex_vertex_index c, hex_vertex_index d,
             hex_facet_index *id);

static void reorder_quad(vertex_index *quad);

static bool hex_compatibility(const vertex_index *a, const vertex_index *b) {
  uint32_t num_shared;
  hex_vertex_index id_a[8], id_b[8];

  num_shared = shared_vertices(a, b, id_a, id_b);

  if (num_shared <= 1)
    return true;
  else if (num_shared == 2) {
    return is_edge(id_a[0], id_a[1]) && is_edge(id_b[0], id_b[1]);
  }
  else if (num_shared == 4) {
    hex_facet_index f_a, f_b;
    if (!is_face(id_a[0], id_a[1], id_a[2], id_a[3], &f_a) ||
        !is_face(id_b[0], id_b[1], id_b[2], id_b[3], &f_b))
      return false;

    static const hex_vertex_index facet[][4] = {
      {0, 1, 2, 3},
      {4, 5, 6, 7},
      {0, 3, 7, 4},
      {1, 2, 6, 5},
      {0, 1, 5, 4},
      {3, 2, 6, 7},
    };

    vertex_index q_a[4] = {a[facet[f_a][0]], a[facet[f_a][1]],
                           a[facet[f_a][2]], a[facet[f_a][3]]};
    vertex_index q_b[4] = {b[facet[f_b][0]], b[facet[f_b][1]],
                           b[facet[f_b][2]], b[facet[f_b][3]]};

    reorder_quad(q_a);
    reorder_quad(q_b);

    return q_a[0] == q_b[0] && q_a[1] == q_b[1] &&
        q_a[2] == q_b[2] && q_a[3] == q_b[3];
  }
  else
    return false;
}

static
hex_vertex_index find_in_hex(const vertex_index *hex, vertex_index v);

static
uint32_t shared_vertices(const vertex_index *a, const vertex_index *b,
                         hex_vertex_index *id_a, hex_vertex_index *id_b) {
  uint32_t num_shared = 0;
  for (hex_vertex_index i = 0; i < 8; i++) {
    vertex_index v = a[i];
    hex_vertex_index j = find_in_hex(b, v);
    id_a[num_shared] = i;
    id_b[num_shared] = j;
    if (j != 8) num_shared++;
  }

  return num_shared;
}

static
hex_vertex_index find_in_hex(const vertex_index *hex, vertex_index v) {
  hex_vertex_index found_i = 9;
  for (hex_vertex_index i = 1; i <= 8; i++) {
    if (hex[i-1] == v) found_i = i;
  }
  return found_i - 1;
}

static
bool is_edge(hex_vertex_index a, hex_vertex_index b) {
  static const hex_vertex_index neighbors[][3] = {
    {1, 3, 4},
    {0, 2, 5},
    {1, 3, 6},
    {0, 2, 7},
    {0, 5, 7},
    {1, 4, 6},
    {2, 5, 7},
    {3, 4, 6},
  };

  return neighbors[a][0] == b || neighbors[a][1] == b || neighbors[a][2] == b;
}

static
bool is_face(hex_vertex_index a, hex_vertex_index b,
             hex_vertex_index c, hex_vertex_index d,
             hex_facet_index *id) {
  static const hex_facet_index shared_facet[8][8][8] = {
    {{0, 0, 0, 0, 2, 4, 6, 2},
     {0, 0, 0, 0, 4, 4, 6, 6},
     {0, 0, 0, 0, 6, 6, 6, 6},
     {0, 0, 0, 0, 2, 6, 6, 2},
     {2, 4, 6, 2, 2, 4, 6, 2},
     {4, 4, 6, 6, 4, 4, 6, 6},
     {6, 6, 6, 6, 6, 6, 6, 6},
     {2, 6, 6, 2, 2, 6, 6, 2},
    },
    {{0, 0, 0, 0, 4, 4, 6, 6},
     {0, 0, 0, 0, 4, 3, 3, 6},
     {0, 0, 0, 0, 6, 3, 3, 6},
     {0, 0, 0, 0, 6, 6, 6, 6},
     {4, 4, 6, 6, 4, 4, 6, 6},
     {4, 3, 3, 6, 4, 3, 3, 6},
     {6, 3, 3, 6, 6, 3, 3, 6},
     {6, 6, 6, 6, 6, 6, 6, 6},
    },
    {{0, 0, 0, 0, 6, 6, 6, 6},
     {0, 0, 0, 0, 6, 3, 3, 6},
     {0, 0, 0, 0, 6, 3, 3, 5},
     {0, 0, 0, 0, 6, 6, 5, 5},
     {6, 6, 6, 6, 6, 6, 6, 6},
     {6, 3, 3, 6, 6, 3, 3, 6},
     {6, 3, 3, 5, 6, 3, 3, 5},
     {6, 6, 5, 5, 6, 6, 5, 5},
    },
    {{0, 0, 0, 0, 2, 6, 6, 2},
     {0, 0, 0, 0, 6, 6, 6, 6},
     {0, 0, 0, 0, 6, 6, 5, 5},
     {0, 0, 0, 0, 2, 6, 5, 2},
     {2, 6, 6, 2, 2, 6, 6, 2},
     {6, 6, 6, 6, 6, 6, 6, 6},
     {6, 6, 5, 5, 6, 6, 5, 5},
     {2, 6, 5, 2, 2, 6, 5, 2},
    },
    {{2, 4, 6, 2, 2, 4, 6, 2},
     {4, 4, 6, 6, 4, 4, 6, 6},
     {6, 6, 6, 6, 6, 6, 6, 6},
     {2, 6, 6, 2, 2, 6, 6, 2},
     {2, 4, 6, 2, 1, 1, 1, 1},
     {4, 4, 6, 6, 1, 1, 1, 1},
     {6, 6, 6, 6, 1, 1, 1, 1},
     {2, 6, 6, 2, 1, 1, 1, 1},
    },
    {{4, 4, 6, 6, 4, 4, 6, 6},
     {4, 3, 3, 6, 4, 3, 3, 6},
     {6, 3, 3, 6, 6, 3, 3, 6},
     {6, 6, 6, 6, 6, 6, 6, 6},
     {4, 4, 6, 6, 1, 1, 1, 1},
     {4, 3, 3, 6, 1, 1, 1, 1},
     {6, 3, 3, 6, 1, 1, 1, 1},
     {6, 6, 6, 6, 1, 1, 1, 1},
    },
    {{6, 6, 6, 6, 6, 6, 6, 6},
     {6, 3, 3, 6, 6, 3, 3, 6},
     {6, 3, 3, 5, 6, 3, 3, 5},
     {6, 6, 5, 5, 6, 6, 5, 5},
     {6, 6, 6, 6, 1, 1, 1, 1},
     {6, 3, 3, 6, 1, 1, 1, 1},
     {6, 3, 3, 5, 1, 1, 1, 1},
     {6, 6, 5, 5, 1, 1, 1, 1},
    },
    {{2, 6, 6, 2, 2, 6, 6, 2},
     {6, 6, 6, 6, 6, 6, 6, 6},
     {6, 6, 5, 5, 6, 6, 5, 5},
     {2, 6, 5, 2, 2, 6, 5, 2},
     {2, 6, 6, 2, 1, 1, 1, 1},
     {6, 6, 6, 6, 1, 1, 1, 1},
     {6, 6, 5, 5, 1, 1, 1, 1},
     {2, 6, 5, 2, 1, 1, 1, 1},
    },
  };

  hex_facet_index f_a = shared_facet[a][b][c];
  hex_facet_index f_b = shared_facet[a][b][d];

  *id = f_a;

  return f_a != 6 && f_a == f_b;
}

static void reorder_quad(vertex_index *quad) {
  quad_vertex_index min_index = 0;
  for (quad_vertex_index i = 0; i < 4; i++) {
    if (quad[i] < quad[min_index]) min_index = i;
  }

  bool positive = quad[(min_index + 1) % 4] < quad[(min_index + 3) % 4];

  vertex_index ordered_quad[4];
  for (quad_vertex_index i = 0; i < 4; i++) {
    if (positive)
      ordered_quad[i] = quad[(min_index + i) % 4];
    else
      ordered_quad[i] = quad[(min_index + (4-i)) % 4];
  }

  memcpy(quad, ordered_quad, sizeof(ordered_quad));
}
