/*
 * topological-hex, a program to compute combinatorial hexahedral meshes.
 *
 * Copyright (C) <2018> <Université catholique de Louvain (UCL), Belgique>
 *
 * List of the contributors to the development, description and complete
 * License: see LICENSE file.
 *
 * This program (topological-hex) is free software:
 * you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see COPYING file).  If not,
 * see <http://www.gnu.org/licenses/>.
 */


#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdbool.h>
#include <limits.h>
#include <string.h>

#include <pthread.h>

#include "types.h"
#include "utils.h"
#include "vertex_set.h"
#include "mesh.h"
#include "solver.h"
#include "dfs.h"
#include "parallel_dfs.h"

static
void on_solution(const solver *solver, uint32_t num_hexes, void *data);

int main(int argc, char **argv) {
  int status = 0;

  uint32_t num_hexes, num_internal_vertices;

  if (argc > 5 || argc < 4) {
    fprintf(stderr, "Usage: %s INPUT H V [SYMMETRIES]\n", argv[0]);
    return 1;
  }

  num_hexes = atoi(argv[2]);
  num_internal_vertices = atoi(argv[3]);

  mesh mesh;
  if (mesh_load(&mesh, argv[1]) != SUCCESS) {
    status = 1;
    fprintf(stderr, "Failed to load mesh from file: %s\n", argv[0]);
    goto fail_mesh_load;
  }

  if (mesh_reserve(&mesh, num_hexes) != SUCCESS) {
    status = 1;
    fprintf(stderr, "Failed to reserve memory for hexes\n");
    goto fail_mesh_reserve;
  }

  solver solver;
  if (solver_init(&solver, &mesh,
                  mesh.num_vertices + num_internal_vertices) != SUCCESS) {
    status = 1;
    fprintf(stderr, "Failed to allocate memory for the solver\n");
    goto fail_solver_init;
  }


  if (argc == 5) {
    vertex_index *symmetries;
    vertex_index num_boundary_vertices;
    size_t num_symmetries;

    if (read_symmetries(argv[4], &symmetries, &num_boundary_vertices,
                        &num_symmetries) == SUCCESS) {
      solver_set_symmetries(&solver, num_symmetries, num_boundary_vertices,
                            symmetries);
    }
    else
      fprintf(stderr, "Failed to load symmetries from file: `%s'\n", argv[4]);
  }

  pthread_mutex_t lock;
  if (pthread_mutex_init(&lock, NULL) != 0) {
    status = 1;
    fprintf(stderr, "Failed to initialize lock!\n");
    goto fail_init_lock;
  }

  if (solver_run_parallel(&solver, on_solution, &lock) != SUCCESS) {
    status = 1;
    fprintf(stderr, "Failed to run solver in parallel\n");
    goto fail_parallel_solve;
  }

  printf("# nodes: %zu\n", solver.num_branches);

fail_parallel_solve:
  pthread_mutex_destroy(&lock);
fail_init_lock:
  solver_release(&solver);
fail_solver_init:
fail_mesh_reserve:
  mesh_release(&mesh);
fail_mesh_load:
  return status;
}

static bool hex_compatibility(const vertex_index *a, const vertex_index *b);

static
void on_solution(const solver *solver, uint32_t num_hexes, void *data) {
  pthread_mutex_t *lock = data;

  pthread_mutex_lock(lock);

  printf("SOLUTION!\n");

  for (hex_index i = 0; i < num_hexes; i++) {
    printf("  %u) {", i);
    for (hex_vertex_index j = 0; j < 8; j++) {
      printf("%u", solver->hexes[8*i+j]);
      if (j + 1 != 8) printf(", ");
    }
    printf("}\n");

    for (hex_index j = i+1; j < num_hexes; j++) {
      if (!hex_compatibility(solver->hexes + 8*i, solver->hexes + 8*j)) {
        printf("  ERROR: Hex #%u and #%u are incompatible!\n", i, j);
      }
    }
  }

  pthread_mutex_unlock(lock);
}

static
uint32_t shared_vertices(const vertex_index *a, const vertex_index *b,
                         hex_vertex_index *id_a, hex_vertex_index *id_b);

static
bool is_edge(hex_vertex_index a, hex_vertex_index b);

static
bool is_face(hex_vertex_index a, hex_vertex_index b,
             hex_vertex_index c, hex_vertex_index d);

static bool hex_compatibility(const vertex_index *a, const vertex_index *b) {
  uint32_t num_shared;
  hex_vertex_index id_a[8], id_b[8];

  num_shared = shared_vertices(a, b, id_a, id_b);

  if (num_shared <= 1)
    return true;
  else if (num_shared == 2) {
    return is_edge(id_a[0], id_a[1]) && is_edge(id_b[0], id_b[1]);
  }
  else if (num_shared == 4) {
    return is_face(id_a[0], id_a[1], id_a[2], id_a[3]) &&
      is_face(id_b[0], id_b[1], id_b[2], id_b[3]);
  }
  else
    return false;
}

static
hex_vertex_index find_in_hex(const vertex_index *hex, vertex_index v);

static
uint32_t shared_vertices(const vertex_index *a, const vertex_index *b,
                         hex_vertex_index *id_a, hex_vertex_index *id_b) {
  uint32_t num_shared = 0;
  for (hex_vertex_index i = 0; i < 8; i++) {
    vertex_index v = a[i];
    hex_vertex_index j = find_in_hex(b, v);
    id_a[num_shared] = i;
    id_b[num_shared] = j;
    if (j != 8) num_shared++;
  }

  return num_shared;
}

static
hex_vertex_index find_in_hex(const vertex_index *hex, vertex_index v) {
  hex_vertex_index found_i = 9;
  for (hex_vertex_index i = 1; i <= 8; i++) {
    if (hex[i-1] == v) found_i = i;
  }
  return found_i - 1;
}

static
bool is_edge(hex_vertex_index a, hex_vertex_index b) {
  static const hex_vertex_index neighbors[][3] = {
    {1, 3, 4},
    {0, 2, 5},
    {1, 3, 6},
    {0, 2, 7},
    {0, 5, 7},
    {1, 4, 6},
    {2, 5, 7},
    {3, 4, 6},
  };

  return neighbors[a][0] == b || neighbors[a][1] == b || neighbors[a][2] == b;
}

static
bool is_face(hex_vertex_index a, hex_vertex_index b,
             hex_vertex_index c, hex_vertex_index d) {
  static const hex_facet_index shared_facet[8][8][8] = {
    {{0, 0, 0, 0, 2, 4, 6, 2},
     {0, 0, 0, 0, 4, 4, 6, 6},
     {0, 0, 0, 0, 6, 6, 6, 6},
     {0, 0, 0, 0, 2, 6, 6, 2},
     {2, 4, 6, 2, 2, 4, 6, 2},
     {4, 4, 6, 6, 4, 4, 6, 6},
     {6, 6, 6, 6, 6, 6, 6, 6},
     {2, 6, 6, 2, 2, 6, 6, 2},
    },
    {{0, 0, 0, 0, 4, 4, 6, 6},
     {0, 0, 0, 0, 4, 3, 3, 6},
     {0, 0, 0, 0, 6, 3, 3, 6},
     {0, 0, 0, 0, 6, 6, 6, 6},
     {4, 4, 6, 6, 4, 4, 6, 6},
     {4, 3, 3, 6, 4, 3, 3, 6},
     {6, 3, 3, 6, 6, 3, 3, 6},
     {6, 6, 6, 6, 6, 6, 6, 6},
    },
    {{0, 0, 0, 0, 6, 6, 6, 6},
     {0, 0, 0, 0, 6, 3, 3, 6},
     {0, 0, 0, 0, 6, 3, 3, 5},
     {0, 0, 0, 0, 6, 6, 5, 5},
     {6, 6, 6, 6, 6, 6, 6, 6},
     {6, 3, 3, 6, 6, 3, 3, 6},
     {6, 3, 3, 5, 6, 3, 3, 5},
     {6, 6, 5, 5, 6, 6, 5, 5},
    },
    {{0, 0, 0, 0, 2, 6, 6, 2},
     {0, 0, 0, 0, 6, 6, 6, 6},
     {0, 0, 0, 0, 6, 6, 5, 5},
     {0, 0, 0, 0, 2, 6, 5, 2},
     {2, 6, 6, 2, 2, 6, 6, 2},
     {6, 6, 6, 6, 6, 6, 6, 6},
     {6, 6, 5, 5, 6, 6, 5, 5},
     {2, 6, 5, 2, 2, 6, 5, 2},
    },
    {{2, 4, 6, 2, 2, 4, 6, 2},
     {4, 4, 6, 6, 4, 4, 6, 6},
     {6, 6, 6, 6, 6, 6, 6, 6},
     {2, 6, 6, 2, 2, 6, 6, 2},
     {2, 4, 6, 2, 1, 1, 1, 1},
     {4, 4, 6, 6, 1, 1, 1, 1},
     {6, 6, 6, 6, 1, 1, 1, 1},
     {2, 6, 6, 2, 1, 1, 1, 1},
    },
    {{4, 4, 6, 6, 4, 4, 6, 6},
     {4, 3, 3, 6, 4, 3, 3, 6},
     {6, 3, 3, 6, 6, 3, 3, 6},
     {6, 6, 6, 6, 6, 6, 6, 6},
     {4, 4, 6, 6, 1, 1, 1, 1},
     {4, 3, 3, 6, 1, 1, 1, 1},
     {6, 3, 3, 6, 1, 1, 1, 1},
     {6, 6, 6, 6, 1, 1, 1, 1},
    },
    {{6, 6, 6, 6, 6, 6, 6, 6},
     {6, 3, 3, 6, 6, 3, 3, 6},
     {6, 3, 3, 5, 6, 3, 3, 5},
     {6, 6, 5, 5, 6, 6, 5, 5},
     {6, 6, 6, 6, 1, 1, 1, 1},
     {6, 3, 3, 6, 1, 1, 1, 1},
     {6, 3, 3, 5, 1, 1, 1, 1},
     {6, 6, 5, 5, 1, 1, 1, 1},
    },
    {{2, 6, 6, 2, 2, 6, 6, 2},
     {6, 6, 6, 6, 6, 6, 6, 6},
     {6, 6, 5, 5, 6, 6, 5, 5},
     {2, 6, 5, 2, 2, 6, 5, 2},
     {2, 6, 6, 2, 1, 1, 1, 1},
     {6, 6, 6, 6, 1, 1, 1, 1},
     {6, 6, 5, 5, 1, 1, 1, 1},
     {2, 6, 5, 2, 1, 1, 1, 1},
    },
  };

  hex_facet_index f_a = shared_facet[a][b][c];
  hex_facet_index f_b = shared_facet[a][b][d];

  return f_a != 6 && f_a == f_b;
}
