/*
 * topological-hex, a program to compute combinatorial hexahedral meshes.
 *
 * Copyright (C) <2018> <Université catholique de Louvain (UCL), Belgique>
 *
 * List of the contributors to the development, description and complete
 * License: see LICENSE file.
 *
 * This program (topological-hex) is free software:
 * you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see COPYING file).  If not,
 * see <http://www.gnu.org/licenses/>.
 */


#include <stdio.h>
#include <stdlib.h>

#include <pthread.h>

#include "mesh2d.h"
#include "solver2d.h"
#include "dfs2d.h"

static
void on_solution(const solver2d *solver, uint32_t num_quads, void *data);

static
error_code read_valence(
  uint32_t *num_vertices, uint32_t **valence, const char *path);

int main(int argc, char **argv) {
  int status = 0;

  uint32_t num_quads, num_internal_vertices;

  if (argc != 5) {
    fprintf(stderr, "Usage: %s INPUT SINGULARITIES F V\n", argv[0]);
    return 1;
  }

  num_quads = atoi(argv[3]);
  num_internal_vertices = atoi(argv[4]);

  mesh2d mesh;
  if (mesh2d_load(&mesh, argv[1]) != SUCCESS) {
    status = 1;
    fprintf(stderr, "Failed to load mesh from file: %s\n", argv[0]);
    goto fail_mesh_load;
  }

  if (mesh2d_reserve(&mesh, num_quads) != SUCCESS) {
    status = 1;
    fprintf(stderr, "Failed to reserve memory for quads\n");
    goto fail_mesh_reserve;
  }

  uint32_t n;
  uint32_t *valence;
  if (read_valence(&n, &valence, argv[2]) != SUCCESS) {
    status = 1;
    fprintf(stderr, "Failed to read valence file.\n");
    goto fail_read_valence;
  }

  if (n != mesh.num_vertices) {
    status = 1;
    fprintf(stderr, "Incorrect number of values in valence file: %u\n", n);
    goto invalid_valence_file;
  }

  solver2d solver;
  if (solver2d_init(&solver, &mesh,
                    mesh.num_vertices + num_internal_vertices,
                    n, valence) != SUCCESS) {
    status = 1;
    fprintf(stderr, "Failed to allocate memory for the solver\n");
    goto fail_solver_init;
  }

  pthread_mutex_t lock;
  if (pthread_mutex_init(&lock, NULL) != 0) {
    status = 1;
    fprintf(stderr, "Failed to initialize lock!\n");
    goto fail_init_lock;
  }

  solver2d_run(&solver, on_solution, &lock);

  printf("# nodes: %zu\n", solver.num_branches);

  pthread_mutex_destroy(&lock);
fail_init_lock:
  solver2d_release(&solver);
fail_solver_init:
invalid_valence_file:
  free(valence);
fail_read_valence:
fail_mesh_reserve:
  mesh2d_release(&mesh);
fail_mesh_load:
  return status;
}

static
void on_solution(const solver2d *solver, uint32_t num_quads, void *data) {
  pthread_mutex_t *lock = data;

  pthread_mutex_lock(lock);

  printf("SOLUTION!\n");

  for (quad_index i = 0; i < num_quads; i++) {
    printf("  %u) {", i);
    for (quad_vertex_index j = 0; j < 4; j++) {
      const quad_vertex_index permutation[] = {0, 1, 3, 2};
      printf("%u", solver->quads[4*i+permutation[j]]);
      if (j + 1 != 4) printf(", ");
    }
    printf("}\n");
  }

  pthread_mutex_unlock(lock);
}

static
error_code read_valence(uint32_t *num_vertices,
                        uint32_t **valence, const char *path) {
  FILE *file = fopen(path, "r");
  if (!file)
    return IO_ERROR;

  error_code status = SUCCESS;

  char buffer[256];
  if (fgets(buffer, sizeof(buffer), file)) {
    char *end = NULL;
    *num_vertices = strtoul(buffer, &end, 10);
    if (!end || (*end != '\0' && *end != '\n'))
      status = PARSE_ERROR;
    else {
      posix_memalign((void**)valence, ALIGNMENT,
                     sizeof(**valence) * (*num_vertices));
    }
  }

  uint32_t i = 0;
  while (status == SUCCESS &&
         fgets(buffer, sizeof(buffer), file) &&
         i < *num_vertices) {
    char *end = NULL;
    (*valence)[i++] = strtoul(buffer, &end, 10);
    if (!end || (*end != '\0' && *end != '\n'))
      return status = PARSE_ERROR;
  }

  if (status == SUCCESS && i != *num_vertices)
    status = PARSE_ERROR;
  else if (status == SUCCESS && !feof(file))
    status = IO_ERROR;

  fclose(file);

  if (status != SUCCESS)
    free(*valence);

  return status;
}
