/*
 * topological-hex, a program to compute combinatorial hexahedral meshes.
 *
 * Copyright (C) <2018> <Université catholique de Louvain (UCL), Belgique>
 *
 * List of the contributors to the development, description and complete
 * License: see LICENSE file.
 *
 * This program (topological-hex) is free software:
 * you can redistribute it and/or modify it under the terms
 * of the GNU General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (see COPYING file).  If not,
 * see <http://www.gnu.org/licenses/>.
 */

#include <check.h>
#include <stdlib.h>
#include <stdbool.h>

#include "mesh2d.h"
#include "input.h"

static bool find_edge(const mesh2d *mesh, const vertex_index *edge) {
  for (quad_index i = 0; i < mesh->num_boundary_edges; i++) {
    uint32_t j;
    for (j = 0; j < 2 && edge[j] == mesh->edges[2*i + j]; j++);
    if (j == 2)
      return true;
  }

  return false;
}

static void check_edges(const mesh2d *mesh, uint32_t n,
                        const vertex_index (*edge)[2]) {
  for (uint32_t i = 0; i < n; i++) {
    ck_assert_msg(find_edge(mesh, edge[i]), "Missing edge: (%u, %u)",
                  edge[i][0], edge[i][1]);
  }
  ck_assert_int_eq(mesh->num_boundary_edges, n);
  ck_assert_int_eq(mesh->num_edges, n);
}

START_TEST(test_load_error) {
  mesh2d mesh;
  ck_assert(mesh2d_load(&mesh, INPUT_DIR "/does_not_exist.mesh") == IO_ERROR);

  mesh2d_release(&mesh);
} END_TEST

START_TEST(test_load_simple) {
  mesh2d mesh;
  ck_assert(mesh2d_load(&mesh, INPUT_DIR "/simple2d.mesh") == SUCCESS);

  ck_assert_int_eq(mesh.num_vertices, 12);
  vertex_index edges[][2] = {
    {0, 1}, {1, 2}, {2, 3}, {3, 4}, {4, 5}, {5, 6},
    {6, 7}, {7, 8}, {8, 9}, {9, 0}
  };
  check_edges(&mesh, sizeof(edges)/sizeof(*edges), edges);

  mesh2d_release(&mesh);
} END_TEST

START_TEST(test_reserve) {
  mesh2d mesh;
  ck_assert(mesh2d_load(&mesh, INPUT_DIR "/simple2d.mesh") == SUCCESS);

  ck_assert_int_eq(mesh.num_boundary_edges, 10);
  vertex_index edge_copy[10*2];
  memcpy(edge_copy, mesh.edges, sizeof(edge_copy));

  ck_assert(mesh2d_reserve(&mesh, 21) == SUCCESS);

  ck_assert(memcmp(edge_copy, mesh.edges, sizeof(edge_copy)) == 0);
  ck_assert_int_eq(mesh.num_boundary_edges, 10);
  ck_assert_int_eq(mesh.num_edges, 47);
  ck_assert_int_eq(mesh.num_quads, 21);

  for (uint32_t i = 2*mesh.num_boundary_edges; i < 2*mesh.num_edges; i++)
    mesh.edges[i] = 5;

  for (uint32_t i = 0; i < 4*mesh.num_quads; i++)
    mesh.quad_edges[i] = 3;

  for (uint32_t i = 0; i < 4*mesh.num_quads; i++)
    mesh.quad_edges[i] = 4;

  mesh2d_release(&mesh);
} END_TEST

Suite *mesh2d_suite(void) {
  Suite *s = suite_create("Mesh2D");

  TCase *tcase = tcase_create("mesh2d_load");
  tcase_add_test(tcase, test_load_error);
  tcase_add_test(tcase, test_load_simple);
  suite_add_tcase(s, tcase);

  tcase = tcase_create("mesh2d_reserve");
  tcase_add_test(tcase, test_reserve);
  suite_add_tcase(s, tcase);

  return s;
}
